"""Declares :class:`Application`."""
import asyncio
import logging
import types
import typing

import ckms
import fastapi
from ckms.jose.models import JSONWebKeySet
from fastapi.exceptions import RequestValidationError
from fastapi.responses import JSONResponse
from h11 import RemoteProtocolError
from unimatrix.exceptions import CanonicalException

from .conf import settings as defaults
from .exceptions import UnprocessableEntity
from .exceptions import TryAgain
from .exceptions import UpstreamServiceNotAvailable
from .exceptions import UpstreamConnectionFailure
from .exceptions import EXCEPTION_RESPONSES
from .ext import ioc
from .handler import BaseRequestHandler
from .types import IRequestHandler


class Application(fastapi.FastAPI):
    logger: logging.Logger = logging.getLogger('uvicorn')

    #: The JSON Web Key Set (JWKS) offered by the server to encrypt
    #: data and verify signatures.
    jwks: JSONWebKeySet

    #: Tags to include in the JWKS.
    jwks_tags: list[str]

    #: The keychain holding the decryption and signing keys
    #: used by the server.
    keychain: ckms.Keychain

    def __init__(
        self,
        settings: types.ModuleType | None = None,
        provider: ioc.Provider | None = None,
        log_exceptions: bool = False,
        jwks: JSONWebKeySet | None = None,
        jwks_tags: list[str] | None = None,
        keychain: ckms.Keychain | None = None,
        *args: typing.Any,
        **kwargs: typing.Any
    ):
        """The :class:`cbra.Application` class integrates :mod:`cbra` with
        the underlying ASGI implementations (:mod:`fastapi` and :mod:`starlette`).

        The `settings` parameter specifies a settings module that is used to
        configure the ASGI application. If `settings` is ``None`` (the default),
        then :mod:`cbra.conf.settings` is used. In normal case this parameter
        is always ``None``, and is only provided with a settings module during
        test cases.

        The `provider` argument specifies the container to use when injecting
        dependencies. If `provider` is ``None`` (the default) then a new
        container is instantiated. In that case, the
        :attr:`cbra.conf.settings.DEPENDENCIES` is also evualated to inject
        dependencies based on the runtime settings. It is assumed that, if the
        dependency injection container is explicitely provided to the
        constructor, it does not need any further configuration.

        With `log_exceptions`, the application may be configured to explicitely
        log :exc:`cbra.exceptions.CanonicalException` instances. The default
        is ``False``.
        """
        self.jwks = jwks or JSONWebKeySet()
        self.jwks_tags = jwks_tags or ['server']
        self.keychain = keychain or ckms.Keychain()
        self.settings = settings or typing.cast(types.ModuleType, defaults)
        on_startup = kwargs.setdefault('on_startup', [])
        on_startup.append(self.on_startup)
        on_shutdown = kwargs.setdefault('on_shutdown', [])
        on_shutdown.append(self._teardown)

        # Setup the dependency injection container. Load from settings only
        # if there was no container provided to the constructor.
        self.container = provider
        if self.container is None:
            self.container = ioc

        self.log_exceptions = log_exceptions
        kwargs.setdefault('docs_url', '/ui')
        kwargs.setdefault('redoc_url', '/docs')
        exception_handlers = kwargs.setdefault('exception_handlers', {})
        exception_handlers.update({
            asyncio.TimeoutError: self.canonical_exception,
            CanonicalException: self.canonical_exception,
            ConnectionError: self.canonical_exception,
            RequestValidationError: self.canonical_exception,
            TimeoutError: self.canonical_exception,
            RemoteProtocolError: self.canonical_exception,
        })
        super().__init__(*args, **kwargs)

    def add(self,
        handler: typing.Union[
            BaseRequestHandler,
            typing.Type[BaseRequestHandler],
            typing.Type[IRequestHandler]
        ],
        *args: typing.Any,
        **kwargs: typing.Any
    ):
        """Add a request handler to the :class:`Application`."""
        handler.add_to_router(app=self, *args, **kwargs)

    def add_api_route(self, *args: typing.Any, **kwargs: typing.Any): # type: ignore
        methods = kwargs.get('methods') or []
        exception_responses = {
            k: v for k, v in dict[int, typing.Any].items(EXCEPTION_RESPONSES)
            if k not in {415, 422}
        }
        if bool(set(methods) & {"POST", "PUT", "PATCH"}):
            exception_responses.update({
                415: EXCEPTION_RESPONSES[415],
                422: EXCEPTION_RESPONSES[422]
            })
        if kwargs.pop('default_error_responses', True):
            if methods != ['OPTIONS']:
                kwargs['responses'] = {
                    **exception_responses,
                    **(kwargs.get('responses') or {})
                }
        return super().add_api_route(*args, **kwargs)

    async def on_startup(self):
        await self._boot()

    async def _boot(self): # pragma: no cover
        self.logger.info("Booting up ASGI application")
        if hasattr(self.settings, 'DEPENDENCIES'):
            await ioc.load(self.settings.DEPENDENCIES, using=self.container)

        await self.keychain.setup()
        if self.jwks_tags:
            self.jwks = JSONWebKeySet.fromtags(self.keychain, self.jwks_tags)
        await self.boot()

    async def boot(self): # pragma: no cover
        """Invoked when the :class:`Application` is starting up. Subclasses
        may override this method to customize the behavior of the
        :class:`Application` during boot.

        Common use cases to override :meth:`boot()` is to establish database
        connections, fetch the public keys of consumed systems, or other
        run configurations that depend on active connections to other services.
        """
        pass

    def configure_query_runners(self, config: dict):
        """Configures and injects the query runners."""
        for alias, runner_class in dict.items(config):
            self.container.provide(f'cbra.runner.{alias}', runner_class)

    async def canonical_exception(
        self,
        request: fastapi.Request,
        exception: BaseException
    ) -> typing.NoReturn:
        """Handles a canonical exception to a standard error message format."""
        if isinstance(exception, ConnectionRefusedError):
            kwargs = {}
            return await self.canonical_exception(
                request,
                UpstreamServiceNotAvailable(**kwargs),
            )
        elif isinstance(
            exception,
            (BrokenPipeError, ConnectionResetError, ConnectionAbortedError)
        ):
            kwargs = {}
            return await self.canonical_exception(
                request,
                UpstreamConnectionFailure(**kwargs),
            )
        #elif isinstance(exception, UnsatisfiedDependency):
        #    return await self.canonical_exception(
        #        request, FeatureNotSupported()
        #    )
        elif isinstance(exception, RequestValidationError):
            return await self.canonical_exception(
                request=request,
                exception=UnprocessableEntity.fromexc(exception)
            )
        elif isinstance(exception, (asyncio.TimeoutError, TimeoutError)):
            #count = int(request.headers.get('X-Retry') or 0)
            return await self.canonical_exception(request, TryAgain(30))
        elif isinstance(exception, CanonicalException):
            status_code = getattr(exception, 'http_status_code', None) or 500
            if self.log_exceptions:
                self.logger.error(
                    "%s (%s): %s",
                    exception.code,
                    exception.http_status_code,
                    exception.message
                )
                exception.log(self.logger.exception)
            response = JSONResponse(
                status_code=status_code,
                content=exception.as_dict(),
                headers=exception.http_headers
            )
            return response
        else:
            raise NotImplementedError

    def get_jwks(self) -> JSONWebKeySet:
        return self.jwks

    def inject(self, name: str, value: object, force: bool = True):
        """Inject a dependency `value` under the name `name` into the
        dependency container.
        """
        return self.container.provide(name, value, force=force)

    async def _teardown(self): # pragma: no cover
        self.logger.info("Tearing down ASGI application")
        await self.teardown()

    async def teardown(self): # pragma: no cover
        """Invoked when the :class:`cbra.Application` is tearing down."""
        pass
