"""Declares :class:`Endpoint`."""
import collections
import inspect
import typing
import types
from inspect import signature
from inspect import Parameter

import fastapi
import fastapi.params
import pydantic
from fastapi.dependencies.utils import get_dependant
from fastapi.dependencies.utils import get_parameterless_sub_dependant
from fastapi.dependencies.utils import solve_dependencies
from fastapi.dependencies.utils import Dependant

from cbra.digests import SHA256Digest
from cbra.digests import SHA512Digest
from cbra.exceptions import AuthenticationRequired
from cbra.exceptions import CanonicalException
from cbra.exceptions import MissingRequestBody
from cbra.exceptions import UnprocessableEntity
from cbra.negotiation import DefaultContentNegotiation
from cbra.parsers import JSONParser
from cbra.parsers import FormParser
from cbra.parsers import YAMLParser
from cbra.renderers import JSONRenderer
from cbra.renderers import YAMLRenderer
from cbra.types import IContentNegotiation
from cbra.types import ICorsPolicy
from cbra.types import IEndpoint
from cbra.types import IParser
from cbra.types import IPrincipal
from cbra.types import IRenderer
from cbra.types import IResponseDigest
from cbra.types import QueryOptions
from cbra.utils import classproperty
from cbra import cors
from .lib.auth import NullPrincipal
from .utils import order_parameters
from .headers import CORS_HEADERS
from .headers import DIGEST_SCHEMA
from .headers import ORIGIN
from .useragent import UserAgent



class Endpoint(IEndpoint):
    """An endpoint is an abstraction that models the end of a message
    channel through which a system can send or receive messages. The
    :class:`Endpoint` provides a low-level interface to implement such
    termina.
    """
    __module__: str = 'cbra'

    #: The list of available digest algorithms.
    digest_algorithms: typing.List[typing.Type[IResponseDigest]] = [
        SHA256Digest,
        SHA512Digest
    ]

    #: The :class:`cbra.lib.auth.IPrincipal` implementation that is used
    #: to authenticate and authorize requests to this endpoint.
    principal: IPrincipal = NullPrincipal()

    #: A :class:`cbra.cors.BaseCorsPolicy` implementation that enforces
    #: the Cross-Origin Resource Sharing (CORS) policy of the server.
    cors_policy: typing.Type[ICorsPolicy] = cors.DefaultPolicy

    #: The default digest algorithm if none was selected.
    default_digest: typing.Type[IResponseDigest] = SHA256Digest

    #: The default response media type, if not specified by the
    #: ``Accept`` request header.
    default_response_type: str = "application/json"

    #: Resolves the content negotiation class for requests.
    negotiation: typing.Type[IContentNegotiation] = DefaultContentNegotiation

    #: The list of :class:`cbra.types.IParser` implementations that are used
    #: to deserialize the request body.
    parsers: typing.List[typing.Type[IParser]] = [
        JSONParser,
        FormParser,
        YAMLParser,
    ]

    #: The list of renderers used to render response bodies.
    renderers: typing.List[typing.Type[IRenderer]] = [
        JSONRenderer,
        YAMLRenderer
    ]

    #: Fallback renderer to create a response if no content type
    #: could be negotiated.
    fallback_renderer: typing.Type[IRenderer] = JSONRenderer

    @classproperty
    def default_response(cls) -> typing.Dict[str, typing.Any]:
        return {
            'description': cls.response_description, # type: ignore
            'content': {
                x.media_type: {} for x in cls.renderers
            },
            'headers': {
                **CORS_HEADERS,
                'Digest': DIGEST_SCHEMA
            }
        }

    @classproperty
    def returns(cls) -> type[pydantic.BaseModel] | types.UnionType | None:
        if cls.response_model:
            response_model = cls.response_model
        else:
            response_model = signature(cls.handle).return_annotation
        if not isinstance(response_model, types.UnionType)\
        and (inspect.isclass(response_model) and not issubclass(response_model, pydantic.BaseModel)):
            response_model = None
        return None

    @classmethod
    def add_to_router(
        cls,
        *,
        app: IEndpoint.RouterType,
        base_path: str = '/',
        method: str | None = None,
        request_handler: typing.Optional[
            typing.Callable[..., IEndpoint.ResponseTypes]
        ] = None,
        **kwargs: typing.Any
    ) -> None:
        method = method or cls.method
        if method is None:
            raise ValueError("The `method` parameter can not be None.")
        if not str.startswith(base_path, '/'):
            raise ValueError("The `base_path` parameter must start with a slash.")
        base_path = str.rstrip(base_path, '/')
        if cls.mount_path:
            base_path = f'{base_path}/{cls.mount_path}'
        base_path = base_path or '/'

        cls = cls.new(
            method=method,
            **kwargs
        )

        kwargs.setdefault('description', cls.description)
        kwargs.setdefault('name', cls.name)
        kwargs.setdefault('openapi_extra', cls.get_openapi_schema())
        kwargs.setdefault('summary', cls.summary)
        tags = kwargs.setdefault('tags', [])
        tags.extend(cls.tags)

        app.add_api_route(
            path=base_path,
            methods=[method],
            endpoint=cls.as_handler(
                method=method,
                path=base_path,
                request_handler=request_handler
            ),
            response_model=cls.returns,
            responses={
                cls.default_response_code: cls.default_response,
                **kwargs.pop('responses', {}),
            },
            status_code=cls.default_response_code,
            **kwargs
        )

    @classmethod
    def as_handler(
        cls,
        method: str,
        path: str,
        request_handler: typing.Optional[
            typing.Callable[..., IEndpoint.ResponseTypes]
        ] = None
    ) -> typing.Coroutine[typing.Any, typing.Any, IEndpoint.ResponseTypes]:
        """Return a callable that invokes the endpoint handler."""
        cors_policy = cls.cors_policy
        negotiation = cls.negotiation
        query_opts = cls.query_model.null
        if method in {"POST", "PUT", "PATCH"}:
            negotiation = cls.negotiation.with_body
        if method == "OPTIONS":
            cors_policy = cors_policy.as_options
        if cls.queryable:
            query_opts = cls.query_model.as_dependant()

        async def handle(
            request: fastapi.Request,
            principal: IPrincipal = fastapi.Depends(cls.principal_factory),
            origin: str = ORIGIN,
            handler: cls = fastapi.Depends(),
            cors: cors_policy = fastapi.Depends(), # type: ignore
            query: QueryOptions = fastapi.Depends(query_opts),
            negotiation: negotiation = fastapi.Depends(), # type: ignore
            useragent: UserAgent = fastapi.Depends(),
            *args: typing.Any,
            **kwargs: typing.Any
        ) -> IEndpoint.ResponseTypes:
            cors: ICorsPolicy = cors
            negotiation: IContentNegotiation = negotiation
            await cors.process_request(request, handler, origin)
            try:
                handler.setup(
                    request=request,
                    principal=principal,
                    path=path,
                    parser=negotiation.select_parser(cls.parsers),
                    renderer=negotiation.select_renderer(
                        renderers=cls.renderers,
                        default=cls.default_renderer
                    ),
                    cors=cors,
                    digest=negotiation.select_digest(
                        cls.digest_algorithms,
                        default=cls.default_digest
                    ),
                    query=query
                )
                response = await handler.dispatch(handler.handle, *args, **kwargs)
            except CanonicalException as exception:
                response = await exception.handle(
                    request=request,
                    handler=handler,
                    renderer=handler.get_renderer()
                )
            else:
                if not isinstance(response, fastapi.Response):
                    response = await handler.render_to_response(
                        content=response
                    )
            assert isinstance(response, fastapi.Response), repr(response) # nosec
            response.headers['Content-Type'] = handler.get_response_encoding()
            await cors.process_response(request, response, handler, origin)
            handler.process_response(request, response)
            return response

        return cls.update_signature(handle, cls.handle)

    @classmethod
    def get_openapi_schema(cls) -> typing.Dict[str, typing.Any]:
        if not cls.needs_body(None) or not hasattr(cls, 'model'):
            return {}

        schema: typing.Dict[str, typing.Any] = {
            'oneOf': []
        }
        for model in typing.get_args(cls.model):
            if not inspect.isclass(model)\
            or not issubclass(model, pydantic.BaseModel):
                raise TypeError(
                    f"{cls.__name__}.model must be a subclass of "
                    "pydantic.BaseModel or subclass thereof."
                )
            schema['oneOf'].append(model.schema())
        if inspect.isclass(cls.model)\
        and issubclass(cls.model, pydantic.BaseModel):
            schema = cls.model.schema()
        return {
            'requestBody': {
                'content': {
                    p.media_type: {
                        'schema': schema
                    } for p in cls.parsers
                }
            }
        }

    @classmethod
    def needs_body(cls, method: typing.Optional[str]) -> bool:
        return (method or cls.method) in cls.body_methods

    @classmethod
    def needs_parameter(cls, p: Parameter) -> bool:
        # Ignore the self parameter and basically anything that is not
        # a fastapi.params.Param instance. Dependencies that are injected
        # through fastapi.Depends and that require request attributes,
        # should be injected in the constructor.
        return not (p.name == 'self' or not isinstance(p.default, fastapi.params.Param))

    @classmethod
    def update_signature(
        cls,
        runner: typing.Any,
        handle: typing.Any
    ) ->  typing.Coroutine[typing.Any, typing.Any, IEndpoint.ResponseTypes]:
        """Update the signature of `runner` to include the dependencies
        declared by `handle`. Remote variable positional and variable
        keyword arguments from the signature. Return `runner` with its
        updated signature.
        """
        sig = signature(runner)
        params = collections.OrderedDict(sig.parameters.items())
        for p in signature(handle).parameters.values():
            if p.name in params: # pragma: no cover
                raise ValueError(f"Duplicate parameter: '{p.name}'.")
            if not cls.needs_parameter(p):
                continue
            params[p.name] = p # pragma: no cover

        for p in cls.get_path_signature():
            params[p.name] = p

        runner.__signature__ = sig.replace(
            parameters=order_parameters([
                p for p in params.values()
                if p.kind not in {
                    Parameter.VAR_POSITIONAL,
                    Parameter.VAR_KEYWORD
                }
            ])
        )
        return runner

    async def authorize(self) -> typing.Optional[typing.NoReturn]:
        """Hook to perform authorization checks for the request. Must raise
        an appropriate exception if the request is not authorized. This
        method may specify dependencies.
        """
        pass

    async def dispatch(
        self,
        handle: typing.Callable[..., IEndpoint.ResponseTypes],
        **kwargs: typing.Any
    ) -> IEndpoint.ResponseTypes:
        if self.require_authentication and not self.is_authenticated():
            raise AuthenticationRequired
        assert self.parser is not None # nosec
        body = await self.parser.parse(self.request)
        dependant = get_dependant(path=self.path, call=handle)
        dependant.dependencies.extend(self.__get_dependants())
        values, errors, tasks, response, _ = await solve_dependencies(
            request=self.request,
            dependant=dependant,
            body=body,
            dependency_overrides_provider=None
        )

        # Collect all errors into an UnprocessableEntity instance.
        # Other errors are fatal and should produce a generic
        # internal server serror.
        for error in errors:
            location = error.loc_tuple()
            if location == ('body',):
                raise MissingRequestBody
            elif isinstance(error.exc, pydantic.ValidationError):
                raise UnprocessableEntity.fromexc(error.exc)
            else:
                raise NotImplementedError

        if tasks or response.status_code:
            raise NotImplementedError(errors, tasks, response)
            if errors:
                exception = errors[0].exc
                print(errors[0])
                if isinstance(exception, pydantic.ValidationError):
                    raise UnprocessableEntity.fromexc(exception)
        return await handle(**values)

    async def handle(self) -> IEndpoint.ResponseTypes:
        raise NotImplementedError

    async def render_to_response(
        self,
        content: typing.Any
    ) -> fastapi.Response:
        renderer = self.get_renderer()
        return fastapi.Response(
            content=renderer.render(content),
            status_code=self.default_response_code
        )

    def get_dependants(self) -> typing.List[typing.Callable[..., typing.Any]]:
        """Return the list of callables that are solved prior to
        invoking the request handler.
        """
        return []

    def get_renderer(self) -> IRenderer:
        return self.renderer or self.fallback_renderer("application/yml;indent=2")

    def get_response_encoding(self) -> str:
        return self.get_renderer().get_response_encoding()

    def is_authenticated(self) -> bool:
        """Return a boolean indicating if the request is authenticated."""
        return self.principal is not None

    def process_response(
        self,
        request: fastapi.Request,
        response: fastapi.Response
    ) -> None:
        """Add headers or apply other transformations to the response.
        It should not be modified after invoking this method.
        """
        # If the method is not HEAD, OPTIONS add the Digest header.
        if self.digest is not None and self.method not in {"HEAD", "OPTIONS"}:
            response.headers['Digest'] = self.digest.calculate(response.body)

    def __get_dependants(self) -> typing.List[Dependant]:
        dependants = [
            self.authorize
        ]
        dependants.extend(self.get_dependants())
        return [
            get_parameterless_sub_dependant(
                depends=fastapi.Depends(func),
                path=self.path
            ) for func in dependants
        ]