# pylint: skip-file
import jwt
import pytest
import pytest_asyncio
from httpx import AsyncClient
from unimatrix.lib import timezone

import cbra
from cbra.lib.auth import NullJWKS
from ..const import GOOGLE_OPENID_SERVER
from ..eventarcreceiver import EventarcReceiver
from ..googlejwtprincipal import GoogleJWTPrincipal


@pytest.fixture
def message(command_message):
    return command_message


@pytest.fixture
def app():
    return cbra.Application()


@pytest.fixture
def jwks():
    return NullJWKS(True)


@pytest.fixture
def token():
    return jwt.encode(
        {
            'iss': GOOGLE_OPENID_SERVER,
            'aud': "http://cbra.internal",
            'iat': 1,
            'exp': timezone.now() + 1000,
            'sub': 1,
            'email': "foo@bar.baz"
        },
        'foo'
    )


@pytest_asyncio.fixture
async def client(app):
    async with AsyncClient(app=app, base_url="http://cbra.internal") as client:
        yield client


@pytest.mark.asyncio
async def test_eventarcreceiver_missing_token_returns_401(app, client, message):
    app.add(EventarcReceiver(members=["foo@bar.baz"]), '/')
    response = await client.post('/', data=message)
    assert response.status_code == 401


@pytest.mark.asyncio
async def test_eventarcreceiver_missing_email_returns_403(app, client, jwks, message):
    app.add(
        EventarcReceiver(
            members=["foo@bar.baz"],
            principal_class=lambda *a, **k: GoogleJWTPrincipal(jwks=jwks)
        ),
        path='/'
    )
    token = jwt.encode(
        {
            'iss': GOOGLE_OPENID_SERVER,
            'aud': "http://cbra.internal",
            'iat': 1,
            'exp': timezone.now() + 1000,
            'sub': 1,
        },
        'foo'
    )
    headers = {
        'Authorization': f"Bearer {token}"
    }
    response = await client.post('/', data=message, headers=headers)
    assert response.status_code == 403, response.text
    dto = response.json()
    assert dto['kind'] == "MissingClaims"


@pytest.mark.asyncio
async def test_eventarcreceiver_missing_sub_returns_403(app, client, jwks, message):
    app.add(
        EventarcReceiver(
            members=["foo@bar.baz"],
            principal_class=lambda *a, **k: GoogleJWTPrincipal(jwks=jwks)
        ),
        path='/'
    )
    token = jwt.encode(
        {
            'iss': GOOGLE_OPENID_SERVER,
            'aud': "http://cbra.internal",
            'iat': 1,
            'exp': timezone.now() + 1000,
            'email': "foo@bar.baz",
        },
        'foo'
    )
    headers = {
        'Authorization': f"Bearer {token}"
    }
    response = await client.post('/', data=message, headers=headers)
    assert response.status_code == 403, response.text
    dto = response.json()
    assert dto['kind'] == "MissingClaims"


@pytest.mark.asyncio
async def test_google_jwt(app, jwks, token):
    headers = {
        'Authorization': f'Bearer {token}'
    }

    @app.post('/')
    async def f(principal=GoogleJWTPrincipal(jwks=jwks)):
        return {
            'iss': principal.iss,
            'aud': principal.aud
        }

    async with AsyncClient(app=app, base_url="http://cbra.internal") as client:
        response = await client.post('/', headers=headers)
        assert response.status_code == 200, response.json()


@pytest.mark.asyncio
async def test_eventarcreceiver_non_whitelisted_email_returns_403(app, client, jwks, message, token):
    app.add(
        EventarcReceiver(
            members=["not@white.listed"],
            principal_class=lambda *a, **k: GoogleJWTPrincipal(jwks=jwks)
        ),
        path='/'
    )
    headers = {
        'Authorization': f"Bearer {token}"
    }
    response = await client.post('/', data=message, headers=headers)
    assert response.status_code == 403, response.text
    dto = response.json()
    assert dto['kind'] == "Forbidden"
