"""Declares :class:`AuthorizationRequestHandler`."""
import datetime
from typing import Any
from typing import Callable
from typing import Coroutine
from typing import Type
from typing import Union

from fastapi import Depends
from fastapi import Response

from cbra.ext import ioc
import fastapi
from .authorizationrequest import AuthorizationRequest
from .authorizationresponse import AuthorizationResponse
from .baseclient import BaseClient
from .iclient import IClient
from .iopenauthorizationserver import IOpenAuthorizationServer
from .preparedauthorizationrequest import PreparedAuthorizationRequest
from .requestcontext import RequestContext
from .requesthandler import RequestHandler


class AuthorizationRequestHandler(RequestHandler):
    __module__: str = 'cbra.ext.oauth2'
    client_class: Type[IClient] = BaseClient

    @classmethod
    def as_handler(
        cls,
        server: IOpenAuthorizationServer,
        **kwargs: Any
    ) -> Callable[..., Coroutine[Any, Any, Union[AuthorizationResponse, Response]]]:
        """Return a request handler that invokes the handler."""
        client = BaseClient.fromquery(server.clients, True)
        request = PreparedAuthorizationRequest.as_dependant(server, client)

        async def f(
            client: IClient = client,
            response_type: str = fastapi.Query(
                default=...,
                title="Response type",
                description=(
                    "Informs the authorization server of the desired grant type"
                ),
                example="code",
                enum=[
                    "code",
                    "code id_token",
                    "code id_token token",
                    "code token",
                    "id_token",
                    "id_token token",
                    "none",
                    "token"
                ]
            ),
            ctx: RequestContext = ioc.instance(RequestContext),
            dto: PreparedAuthorizationRequest = request,
            handler: AuthorizationRequestHandler = ioc.instance(cls)
        ) -> Union[AuthorizationResponse, Response]:
            handler.setup(**kwargs)
            handler.now = datetime.datetime.utcnow()
            return await handler.handle(
                client=client,
                ctx=ctx,
                dto=dto
            )
        f.__doc__ = cls.__doc__
        return f

    async def handle(
        self,
        client: IClient,
        ctx: RequestContext,
        dto: PreparedAuthorizationRequest
    ) -> Response:
        """Handle an **OAuth 2.0 Authorization Request**."""
        raise NotImplementedError