"""Declares :class:`ClientConfig`."""
import datetime
import secrets
import typing

import ckms
from ckms.jose import JSONWebKey
from ckms.jose import JSONWebKeySet
from ckms.jose import JSONWebToken
from ckms.jose import PayloadCodec
from ckms.jose.exceptions import JOSEException

from .baseclient import BaseClient
from .exceptions import MissingRedirectURL
from .exceptions import RedirectForbidden
from .isubject import ISubject
from .granttype import GrantType
from .redirecturl import RedirectURL


class ClientConfig(BaseClient):
    client_id: str

    client_secret: typing.Optional[str] = None

    redirect_uris: typing.List[RedirectURL] = []

    default_redirect_url: typing.Optional[RedirectURL] = None

    keys: typing.List[JSONWebKey] = []

    grant_types_supported: typing.Optional[typing.List[GrantType]] = []

    allowed_audience: typing.Set[str] = set()

    @property
    def jwks(self) -> JSONWebKeySet:
        obj = JSONWebKeySet()
        for k in self.keys:
            obj.add(k)
        return obj

    def __init__(self, **kwargs: typing.Any):
        redirect_uris = kwargs.setdefault('redirect_uris', [])
        if 'default_redirect_url' in kwargs:
            redirect_uris.append(kwargs['default_redirect_url'])
        super().__init__(**kwargs)

    def allows_audience(self, issuer: str, audience: typing.Set[str]) -> bool:
        """Return a boolean indicating if the client allows the issueing
        of access token for the given audiences set `audience`.
        """
        return (audience <= (self.allowed_audience | {issuer}))

    def allows_grant_type(self, grant_type: GrantType) -> bool:
        """Return a boolean indicating if the client allows the
        given grant type.
        """
        return grant_type.value in self.grant_types_supported

    def can_redirect(self, url: typing.Union[str, RedirectURL]) -> bool:
        """Return a boolean indicating if the client allows
        redirects to the given `url`.
        """
        return url in self.redirect_uris

    async def issue_token(
        self,
        codec: PayloadCodec,
        algorithm: str,
        using: str,
        issuer: str,
        audience: typing.Union[str, typing.Set[str]],
        subject: ISubject,
        ttl: int,
        scope: typing.Set[str]
    ) -> str:
        """Issue an access token encoded as a JSON Web Token (JWT)
        as defined in :rfc:`9068`. Return a string containing the
        token.
        """
        claims = {
            'aud': list(audience),
            'client_id': self.client_id,
            'iss': issuer,
            'jti': secrets.token_urlsafe(24),
            'scope': str.join(' ', sorted(scope)),
            'sub': subject.sub,
        }
        now = claims['iat'] = int(datetime.datetime.utcnow().timestamp())
        claims.update({
            'exp': now + ttl,
            'nbf': now,
        })
        token = await codec.encode(JSONWebToken(**claims))\
            .sign(algorithm=algorithm, using=using, typ="at+jwt")
        return bytes.decode(bytes(token), 'ascii')

    def is_public(self) -> bool:
        """Return a boolean indicating if the client is public."""
        return self.client_secret is None

    def get_redirect_url(
        self,
        url: typing.Optional[typing.Union[str, RedirectURL]] = None
    ) -> RedirectURL:
        """Return a :class:`RedirectURL` instance that the server is
        allowed to redirect to.
        """
        if not isinstance(url, RedirectURL):
            url = RedirectURL.validate(url)
        url = url or self.default_redirect_url
        if url is None:
            raise MissingRedirectURL
        if not self.can_redirect(url):
            raise RedirectForbidden
        return url

    async def verify_jwt(
        self,
        token: typing.Union[bytes, str],
        decrypter: typing.Optional[ckms.BaseKeySet],
        compact_only: bool = False,
    ) -> typing.Dict[str, typing.Any]:
        """Verifies a JSON Web Token (JWT) using the public keys of the
        client.

        Args:
            token (byte-sequence or string): the JWT to deserialize and
                verify.
            decrypter: a :class:`~ckms.BaseKeySet` implementation that has the
                private keys to decrypt, if the payload is a JWE.
            compact_only (bool): indicates if only compact deserialization
                is allowed. If `compact_only` is ``True`` and the token
                is not compact serialized, an exception is raised.

        Returns
            A :class:`dict` containing the claims in the JWT.
        """
        codec = PayloadCodec(decrypter=decrypter, verifier=self.jwks)
        if isinstance(token, str): # pragma: no cover
            token = str.encode(token, 'ascii')
        try:
            jwt = await codec.decode(token)
        except (JOSEException, ValueError) as e:
            raise JOSEException from e
        return dict(jwt)
