"""Declares :class:`ISubject`."""
import abc

import typing

from ckms.jose.models import JSONWebKey
from ckms.jose import JSONWebSignature


class ISubject(metaclass=abc.ABCMeta):
    """Specifies the interface to which all subject implementations
    must provide.
    """
    __module__: str = 'cbra.ext.oauth2'

    #: The subject identifier. The type and semantics are implementation specific,
    #: but the concrete implementation must ensure that :attr:`sub` is JSON
    #: serializable.
    sub: typing.Union[bytes, int, str]

    #: Identifies the client in which the context subject was
    #: instantiated, or ``None``.
    client_id: typing.Optional[str] = None

    #: Used for testing purposes only.
    client_scope: typing.Dict[str, typing.Set[str]]

    def allows_scope(self, scope: typing.Set[str]) -> bool:
        """Return a boolean indicating that the requested scope `scope`
        was allowed by the subject.
        """
        raise NotImplementedError

    def register_public_key(self, key: JSONWebKey) -> None:
        """Register a public key for the :class:`ISubject`."""
        raise NotImplementedError

    def verify(
        self,
        jws: JSONWebSignature
    ) -> typing.Union[bool, typing.Coroutine[typing.Any, typing.Any, bool]]:
        """Verifies a JSON Web Signature (JWS) using the keys that were
        previously registered by the :term:`Subject`.
        """
        raise NotImplementedError