"""Declares :class:`TokenIssuer`."""
import typing

from ckms.jose import ClaimSet
from ckms.jose import PayloadCodec
from ckms.jose import Malformed
from ckms.jose import InvalidClaim
from ckms.jose import InvalidToken

from .exceptions import InvalidGrant
from .exceptions import InvalidScope
from .exceptions import AssertionReplayed
from .iclient import IClient
from .istorage import IStorage
from .isubjectrepository import ISubjectRepository
from .granttype import GrantType
from .params import CurrentServerMetadata
from .params import LocalIssuer
from .params import ServerCodec
from .params import SubjectRepository
from .params import TransientStorage
from .utils import dispatcher
from .servermetadata import ServerMetadata
from .tokenrequest import TokenRequest
from .tokenrequestclient import TokenRequestClient
from .tokenresponse import TokenResponse
from .utils import classproperty


class TokenIssuer:
    """Issues a token for various grant types."""
    __module__: str = 'cbra.ext.oauth2'

    #: Specifies the maximum age of an assertion.
    max_assertion_age: int = 30000

    #: Specifies the time-to-live of an access token if not otherwise
    #: specified, in seconds.
    default_ttl: int = 300

    #: The client that is handling the token request.
    client: IClient

    #: The concrete :class:`~cbra.ext.oauth2.ISubjectRepository`
    #: implementation that is used to lookup :term:`Subjects`.
    subjects: ISubjectRepository

    #: A :class:`~ckms.ext.oauth2.ServerMetadata` instance describing
    #: the authorization server that is issueing a token.
    metadata: ServerMetadata

    #: A :class:`ckms.jose.PayloadCodec` instance that knows how to
    #: decrypt JWE objects that were encrypted using the keys exposed
    #: at the JWKS URL.
    codec: PayloadCodec

    #: Indicates if a lookup of the JWKS of any issuer is allowed.
    allow_jwks_lookups: bool = True

    #: The set of trusted issuers for which the keys may be retrieved.
    trusted_issuers: typing.Set[str] = set()

    #: The URL of the local issuer i.e. the server. It is assumed that
    #: all tokens issued by this server have this value set to their
    #: issuer attribute.
    issuer: str

    #: The transient storage used.
    storage: IStorage

    @classproperty
    def grant_types_supported(cls) -> typing.List[str]:
        return list(cls.grant.handlers.keys())

    def __init__(
        self,
        client: IClient = TokenRequestClient,
        subjects: ISubjectRepository = SubjectRepository,
        codec: PayloadCodec = ServerCodec,
        metadata: ServerMetadata = CurrentServerMetadata,
        issuer: str = LocalIssuer,
        storage: IStorage = TransientStorage
    ):
        self.client = client
        self.subjects = subjects
        self.codec = codec
        self.metadata = metadata
        self.issuer = issuer
        self.storage = storage

    def is_local_issued(self, claims: ClaimSet) -> bool:
        """Return a boolean indicating if the claimset was issued
        by the local server.
        """
        return self.issuer == claims.iss

    @dispatcher
    async def grant(
        self,
        dto: TokenRequest
    ) -> typing.Tuple[str, TokenRequest]:
        """Dispatches a request to the **Token Endpoint** to the
        appropriate handler for its ``grant_type``.
        """
        return (str(dto.grant_type.value), dto)

    @grant.on(GrantType.jwt_bearer.value) # type: ignore
    @InvalidGrant.catch(Malformed, "The `assertion` parameter is malformed.")
    @InvalidGrant.catch(InvalidClaim)
    @InvalidGrant.catch(InvalidToken)
    async def grant_jwt_bearer(
        self,
        dto: TokenRequest
    ) -> TokenResponse:
        jws = await self.codec.decode_jws(dto.assertion)
        claims = jws.get_claimset()
        claims.verify(
            audience={self.metadata.token_endpoint},
            max_age=self.max_assertion_age,
            required={'iss', 'aud', 'sub', 'iat', 'exp', 'nbf', 'jti'}
        )

        # Check if the assertion is not being replayed.
        if await self.storage.consume(claims):
            raise AssertionReplayed

        # There are two cases that need to be handled here, either the
        # claims are self-signed (a public key is preregistered for the
        # subject specified in the `sub` claim), or the claims were
        # signed by a trusted third-party.
        subject = await self.subjects.get(
            client_id=self.client.client_id,
            subject_id=typing.cast(typing.Union[int, str], claims.sub)
        )
        if not subject.allows_scope(dto.scope):
            raise InvalidScope

        # Lookup the JWKS of the issuer as indicated by the `iss` claim,
        # or use the keys pre-registered by the resource owner if the claims
        # are self-signed.
        selfsigned = claims.is_selfsigned()
        if not selfsigned and not self.is_local_issued(claims):
            # The assertion was issued and signed by a third-party.
            if not self.allow_jwks_lookups:
                raise InvalidGrant(
                    error_description=(
                        "The assertion was signed by a third-party, but it's "
                        "public keys are not known to the server and it does "
                        "not allow import using the JWKS URI."
                    )
                )
            if claims.iss not in self.trusted_issuers:
                raise InvalidGrant(
                    error_description=(
                        "The assertion was issued and signed by an issuer that is "
                        "not trusted by the server."
                    )
                )
            raise NotImplementedError
        elif not selfsigned:
            # The assertion was issued and signed by this server.
            raise InvalidGrant(
                error_description=(
                    "The assertion was issued and signed by an issuer that is "
                    "not trusted by the server."
                )
            )
        else:
            verifier = subject

        if not (await verifier.verify(jws)):
            raise InvalidGrant(
                error_description=(
                    "The signature of the JWT assertion did not validate. Make "
                    "sure that the signing keys are pre-registered by the "
                    "resource owner."
                )
            )

        # If the request does not include a "resource" parameter, the authorization
        # server MUST use a default resource indicator in the "aud" claim (RFC 9068,
        # Section 3).
        dto.resource.add(self.issuer)
        if not self.client.allows_audience(self.issuer, dto.resource):
            raise InvalidGrant(
                error_description=(
                    "Any or all of the audiences are not allowed by the client."
                )
            )

        return TokenResponse(
            access_token=await self.client.issue_token(
                codec=self.codec,
                algorithm='EdDSA',
                using='sig',
                issuer=self.issuer,
                audience=dto.resource,
                subject=subject,
                ttl=self.default_ttl,
                scope=set(dto.scope)
            ),
            token_type="Bearer",
            expires_in=self.default_ttl
        )