"""Declares :class:`TokenRequest`."""
import typing

import pydantic

from .bodydependantmodel import BodyDependantModel
from .fields import ResourceField
from .granttype import GrantType
from .servermetadata import ServerMetadata
from .spaceseparatedlist import SpaceSeparatedList
from .stringorlist import StringOrList


class TokenRequest(BodyDependantModel):
    """The parameters available for requests to the **Token Endpoint**."""
    __module__: str = 'cbra.ext.oauth2'

    client_id: typing.Optional[str]

    grant_type: GrantType = pydantic.Field(
        default=...,
        title="Grant type",
        description=(
            "Specifies the grant type that is requested by the client."
        ),
        example="authorization_code"
    )

    scope: typing.Optional[SpaceSeparatedList] = pydantic.Field(
        default=[],
        title="Scope",
        description=(
            "The requested scope that was previously granted to the subject "
            "or client."
        )
    )

    assertion: str = pydantic.Field(
        default=None,
        title="Assertion",
        description=(
            "The digitally signed document asserting the identity of the "
            "subject.\n\n"
            f"For grants of type `{GrantType.jwt_bearer.value}`, "
            "the `assertion` parameter contains a single JWT. The claims "
            "`exp`, `iat`, `nbf` must be present, in addition to the following"
            " claims:"
            "\n- `iss` - the OAuth 2.0 client id if the assertion is self-issued, "
            "otherwise the identifier of the third-party issuer."
            "\n- `aud` - the URL of the token endpoint."
            "\n- `sub` - identifies the subject on behalf of whom the access token "
            "is requested, i.e. the resource owner."
            "\n- `jti` - a randomly generated identifier for this specific assertion."
        )
    )

    resource: typing.Optional[StringOrList] = ResourceField()

    def is_asserted(self) -> bool:
        """Return a boolean indicating if the request is asserted as defined
        in :rfc:`7521` **Assertion Framework for OAuth 2.0 Client
        Authentication and Authorization Grants**.
        """
        return self.assertion is not None
