"""Declares :func:`TokenRequestClient`."""
import functools
import typing

import fastapi
from fastapi.security.utils import get_authorization_scheme_param

from .exceptions import ClientAuthenticationRequired
from .exceptions import InvalidClient
from .exceptions import InvalidClientAuthenticationScheme
from .exceptions import InvalidGrant
from .exceptions import NoSuchClient
from .fields import GrantTypeField
from .granttype import GrantType
from .iclient import IClient
from .iclientrepository import IClientRepository
from .params import ClientRepository


__all__ = ['TokenRequestClient']




async def get_token_request_client(
    authorization: typing.Optional[str] = fastapi.Header(
        default=None,
        alias='Authorization'
    ),
    client_id: typing.Optional[str] = fastapi.Body(None, embed=True),
    grant_type: GrantType = GrantTypeField(
        functools.partial(fastapi.Body, embed=True)
    ),
    clients: IClientRepository = ClientRepository
) -> typing.Optional[IClient]:
    """Retrieve the current client from either the ``client_id`` specified
    in the request body or the ``Authorization`` header (for confidential
    clients).
    """
    # Requests to the Token Endpoint always have a client specified. If either
    # the header or the client_id in the request body is missing, we are in an
    # error state.
    if not authorization and not client_id:
        raise InvalidClient(
            error="invalid_client",
            error_description=(
                "Specify the `client_id` parameter in the request body if "
                "the client is public, else provide credentials with the "
                "`Authorization` header."
            )
        )
    if (client_id is not None and not await clients.exists(client_id)):
        raise NoSuchClient

    # If the client attempted to authenticate via the "Authorization" 
    # request header field, the authorization server MUST respond with
    # an HTTP 401 (Unauthorized) status code and include the
    # "WWW-Authenticate" response header field matching the
    # authentication scheme used by the client (RFC 6749, Section 5.2).
    client = None
    scheme, credentials = get_authorization_scheme_param(authorization)
    if scheme and credentials:
        if str.lower(scheme) != "bearer":
            raise InvalidClientAuthenticationScheme(
                using=scheme,
                required="Bearer"
            )
        raise NotImplementedError
        # client_id = await clients.fromcredentials(credentials.credentials)

    # If the client_id variable is None at this point, then it was not provided
    # in the request body, could not be determined from the Authorization header,
    # and there was no client assertion or the client assertion did not verify.
    if client_id is None: # pragma: no cover
        raise NoSuchClient

    # A client MAY use the "client_id" request parameter to identify itself
    # when sending requests to the token endpoint.  In the "authorization_code"
    # "grant_type" request to the token endpoint, an unauthenticated client
    # MUST send its "client_id" to prevent itself from inadvertently accepting
    # a code intended for a client with a different "client_id".  This protects
    # the client from substitution of the authentication code
    # (RFC 6749, Section 3.2.1).
    client = await clients.get(client_id)

    # Ensure that we are dealing with a public client.
    if not client.is_public():
        raise ClientAuthenticationRequired

    # Ensure that the client allows the requested grant.
    if not client.allows_grant_type(grant_type):
        raise InvalidGrant(
            error_description=(
                "The client does not allow the given grant: "
                f"{grant_type.value}."
            )
        )

    return client


TokenRequestClient = fastapi.Depends(get_token_request_client)