# pylint: skip-file
"""Webhooks are an integral part of modern web applications. Many applications
consume external services through (REStful) APIs, and those services often
provide a mechanism to deliver event notifications to their consumers: webhooks.
The :mod:`cbra` framework provides a class-oriented solution to handle these
incoming messages.


Basic example
=============
The most simple implementation of a webhook handler is the example below:

.. literalinclude:: ../../../examples/webhook-simple.py
  :language: python
  :caption: webhook.py
  :lines: 2,4-13


Message authentication
======================
There is an obvious security issue if the webhook would accept just any
message without verifying its contents. As many webhook implementations
use Hash-Based Message Authentication Code (HMAC) using the SHA-256
algorithm to prove the authenticity and integrity of the message,
:class:`~cbra.ext.webhook.Webhook` provides a set of handy hooks to verify
this. Expanding on the previous example:

.. literalinclude:: ../../../examples/webhook-simple.py
  :language: python
  :caption: webhook.py
  :lines: 2-6,8-10,15-19

.. note::

  HTTP headers that are not in the HTTP standards should be prefixed
  with ``X-``, to prevent clashes with future standards.

The example above may be tested with the following values:

- ``X-Signature``: ``b247d1f8d1d1ce1a5755308b10c9bbed4b092200d4f14d2fb3806b78927082be``
- Request body: ``{"event": "greeting", "message": "Hello world!"}``

Consider that the :meth:`~cbra.ext.webhook.Webhook.get_secret()` method is
asynchronous, because retrieving the secret might involve a call to a remote
source, such as a database or cloud secret storage.

**The default implementation computes the HMAC over the request body.**
If you need to implement different message verification algorithms or
strategies, the :class:`~cbra.ext.webhook.Webhook` subclass may override the
:meth:`~cbra.ext.webhook.Webhook.verify_signature()` method


Additional resources
====================


Full example
============
.. literalinclude:: ../../../examples/webhook-simple.py
  :language: python
  :caption: webhook.py
  :lines: 2-
"""
from .webhookhandler import WebhookHandler


Webhook = WebhookHandler


__all__ = [
    'Webhook'
]
