# pylint: skip-file
import fastapi
import pytest
import pytest_asyncio
from httpx import AsyncClient

import cbra
from cbra.ext import webhook


class DefaultHandler(webhook.Webhook):

    async def handle(self):
        pass


class ProtectedHandler(webhook.Webhook):
    signature_header = 'X-Signature'

    async def handle(self):
        pass

    async def get_secret(self, request: cbra.Request, **kwargs) -> bytes:
        """Return the secret that is used to verify the message HMAC."""
        return b'secret'


class Base64ProtectedHandler(ProtectedHandler):
    signature_encoding = 'base64'


class RestrictedEventHandler(webhook.Webhook):
    allowed_events = {"event.foo"}

    async def handle(self):
        pass

    def get_event_name(self, dto: dict = fastapi.Body(...)) -> str:
        """Return the name of the webhook event."""
        return dto.get('event')


class DynamicHandler(webhook.Webhook):
    is_dynamic = True

    async def handle(self, webhook_id: int):
        pass


@pytest.fixture
def app():
    app = cbra.Application()
    app.add(DefaultHandler(), path='/default')
    app.add(ProtectedHandler(), path='/protected')
    app.add(Base64ProtectedHandler(), path='/protected/base64')
    app.add(RestrictedEventHandler(), path='/events')
    app.add(DynamicHandler(), path='/dynamic/{webhook_id}')
    return app


@pytest_asyncio.fixture
async def client(app):
    async with AsyncClient(app=app, base_url="http://test") as client:
        yield client
