# pylint: skip-file
import base64
import hmac
import json

import pytest


@pytest.mark.asyncio
async def test_message_is_handled(client):
    response = await client.post(
        url='/default',
        json={}
    )
    assert response.status_code == 200


@pytest.mark.asyncio
async def test_unsigned_message_is_rejected(client):
    response = await client.post(
        url='/protected',
        json={}
    )
    assert response.status_code == 403


@pytest.mark.asyncio
async def test_signed_message_is_accepted_with_valid_signature(client):
    data = json.dumps({})
    signature = hmac.digest(b'secret', str.encode(data), 'sha256')
    response = await client.post(
        url='/protected',
        data=data,
        headers={
            'Content-Type': "application/json",
            'X-Signature': bytes.hex(signature)
        }
    )
    assert response.status_code == 200


@pytest.mark.asyncio
async def test_signed_message_is_rejected_with_invalid_signature(client):
    data = json.dumps({})
    signature = hmac.digest(b'wrong-secret', str.encode(data), 'sha256')
    response = await client.post(
        url='/protected',
        data=data,
        headers={
            'Content-Type': "application/json",
            'X-Signature': bytes.hex(signature)
        }
    )
    assert response.status_code == 403


@pytest.mark.asyncio
async def test_signed_message_is_rejected_with_invalid_encoding(client):
    signature = 'HaHaHaH'
    response = await client.post(
        url='/protected',
        data='{}',
        headers={
            'Content-Type': "application/json",
            'X-Signature': signature
        }
    )
    assert response.status_code == 403


@pytest.mark.asyncio
async def test_signed_message_is_accepted_with_valid_signature_b64(client):
    data = json.dumps({})
    signature = hmac.digest(b'secret', str.encode(data), 'sha256')
    response = await client.post(
        url='/protected/base64',
        data=data,
        headers={
            'Content-Type': "application/json",
            'X-Signature': base64.b64encode(signature)
        }
    )
    assert response.status_code == 200


@pytest.mark.asyncio
async def test_valid_event_type_is_accepted(client):
    response = await client.post(
        url='/events',
        json={
            'event': "event.foo"
        }
    )
    assert response.status_code == 200, response.text


@pytest.mark.asyncio
async def test_invalid_event_type_is_rejected(client):
    response = await client.post(
        url='/events',
        json={
            'event': "event.bar"
        }
    )
    assert response.status_code == 403, response.text


@pytest.mark.asyncio
async def test_dynamic_receives_webhook_id(client):
    response = await client.post(
        url='/dynamic/1',
        json={
            'event': "event.bar"
        }
    )
    assert response.status_code == 200, response.text
