"""Declares :class:`JSONRenderer`."""
import json
import typing

from cbra.types import IRenderer
from cbra.utils import parse_header


class JSONRenderer(IRenderer):
    """
    Renderer which serializes to JSON.
    """
    __module__: str = 'cbra.renderers'
    charset: typing.Optional[str] = None
    compact: bool = True
    #encoder_class = encoders.JSONEncoder
    ensure_ascii: bool = True
    format: str = 'json'
    indent: typing.Optional[int] = None
    media_type: str = 'application/json'
    response_media_type: str = "application/json"
    strict: bool = True
    structured: bool = True

    def get_indent(
        self,
        accepted_media_type: str,
        renderer_context: typing.Dict[str, typing.Any]
    ) -> typing.Optional[int]:
        if self.accepted:
            _, params = parse_header(self.accepted.encode('ascii'))
            try:
                indent = typing.cast(int, params['indent']) # type: ignore
                v = max(min(int(indent), 8), 0)
                return None if v == 0 else v
            except (KeyError, ValueError, TypeError):
                pass

        # If 'indent' is provided in the context, then pretty print the result.
        return renderer_context.get('indent', None) or self.indent

    def render(self, data, renderer_context=None):
        """
        Render `data` into JSON, returning a bytestring.
        """
        if data is None:
            return b''

        renderer_context = renderer_context or {}
        indent = self.get_indent(self.accepted, renderer_context)

        if indent is None:
            separators = (',', ':')
        else:
            separators = (', ', ': ')

        ret = json.dumps(
            data, #cls=self.encoder_class,
            indent=indent, ensure_ascii=self.ensure_ascii,
            allow_nan=not self.strict, separators=separators
        )

        # We always fully escape \u2028 and \u2029 to ensure we output JSON
        # that is a strict javascript subset.
        # See: https://gist.github.com/damncabbage/623b879af56f850a6ddc
        ret = ret.replace('\u2028', '\\u2028').replace('\u2029', '\\u2029')
        return ret.encode()