"""Declares :class:`ResourceMetaclass`."""
from multiprocessing.sharedctypes import Value
import typing
from cbra.resource.iresource import IResource

import fastapi
import fastapi.params
import inflect

from cbra.types import QueryOptions
from .pathparameter import PathParameter


engine = inflect.engine()
VOWEL_CHARS = {'a', 'e', 'i', 'o', 'u'}


class ResourceMetaclass(type):

    @staticmethod
    def get_article(word: str):
        return 'a' if str.lower(word[0]) not in VOWEL_CHARS else 'an'

    def __new__(
        cls,
        name: str,
        bases: typing.Tuple[typing.Type[type], ...],
        namespace: typing.Dict[str, typing.Any]
    ):
        super_new = super().__new__
        if namespace.get('__abstract__', False):
            return super_new(cls, name, bases, namespace)
        name = namespace.setdefault('name', str.replace(name, 'Resource', ''))
        namespace.setdefault('name_article', cls.get_article(name))
        namespace.setdefault('pluralname', engine.plural(name))

        # Collect the attributes from the bases.
        base_attrs = {}
        for b in reversed(bases):
            base_attrs.update(b.__dict__)

        # If resource class is a subresource, then a parent member is
        # contained by the namespace.
        parent: ResourceMetaclass = namespace.get(
            'parent',
            base_attrs.get('parent')
        )

        # Construct the path parameters, including those defined by
        # the parent classes.
        if 'path_parameter' in namespace:
            path_parameter = namespace['path_parameter'] = PathParameter(
                name=namespace['path_parameter'],
                annotation=namespace.pop('path_parameter_class', None)
            )
        else:
            path_parameter = base_attrs.get('path_parameter')

        # Construct a pydantic model holding the query parameters.
        if 'query_parameters' in namespace:
            model_attrs: typing.Dict[typing.Any, typing.Any] = {}
            model_hints: typing.Dict[typing.Any, typing.Any] = {}
            for i, param in enumerate(namespace.pop('query_parameters', [])):
                if param.alias is None:
                    raise ValueError(
                        f"{name}.query_parameters[{i}] must set the "
                        "'alias' parameter."
                    )
                if QueryOptions.is_reserved_field(param.alias):
                    raise ValueError(f"The name '{param.alias}' is reserved.")
                if not isinstance(param, fastapi.params.Query):
                    raise NotImplementedError
                model_attrs[param.alias] = param
                model_hints[param.alias] = param.extra.get('annotation') or str
            if model_attrs:
                namespace['query_model'] = type(
                    f'{name}Query',
                    (QueryOptions,),
                    {**model_attrs, '__annotations__': model_hints}
                )

        # Create the new class with the updated namespace.
        new_class: type[IResource] = typing.cast(
            type[IResource],
            super_new(cls, name, bases, {
                **namespace,
            })
        )

        # The path parameter is required to determine the path for the
        # collection endpoints.
        if new_class.subresources and not new_class.path_parameter:
            raise ValueError(
                f"{name}.path_parameter can not be None because "
                f"{name}.subresources is a non-empty list."
            )

        # Set the parent attribute on the subresources. This is used to
        # determine the path parameters.
        for i, child in enumerate(new_class.subresources):
            new_class.subresources[i] = typing.cast(
                type[IResource],
                child.new(parent=new_class)
            )

        return new_class