"""Declares :class:`Transaction`."""
import uuid

import aorta
from cbra.ext import ioc


class Transaction:
    """Encapsulates a transaction for :mod:`cbra` components, such as
    HTTP request handlers, webhooks, event listeners and commands handlers. The
    :class:`Transaction` object ensures that either all commands, events or
    tasks are published, or none of them.
    """

    @property
    def correlation_id(self) -> uuid.UUID:
        return self._correlation_id

    @property
    def _publisher(self) -> aorta.MessagePublisher:
        return ioc.resolve('MessagePublisher')

    def __init__(self, correlation_id: str = None, *args, **kwargs):
        self._correlation_id = correlation_id or uuid.uuid4()
        self._commands = []
        self._events = []

    async def commit(self):
        """Submits pending commands and events."""
        if self._commands or self._events:
            await self._publisher.publish(
                objects=self._commands + self._events,
                correlation_id=self._correlation_id
            )

    def issue(self, command: aorta.Command):
        """Issue a command using the default command issuer."""
        self._commands.append(command)

    def publish(self, event: aorta.Event):
        """Publish an event using the default event publisher."""
        self._events.append(event)

    async def __aenter__(self):
        return self

    async def __aexit__(self, cls, exception, traceback) -> bool:
        if exception is not None:
            return
        await self.commit()
