"""Example webhook implementation for WooCommerce webhook
event messages.
"""
import enum
import functools
import types
import typing

import fastapi
import pydantic

import cbra
import cbra.parsers
import cbra.renderers


class WooCommercePing(pydantic.BaseModel):
    webhook_id: int


class WooCommerceObject(pydantic.BaseModel):
    id: int = pydantic.Field(
        default=...,
        title="Identifier",
        description=(
            "Unique identifier for the object."
        )
    )


class WooCommerceOrder(WooCommerceObject):
    order_key: str | None


class WooCommerceTopicType(str, enum.Enum):
    coupon_created   = f"coupon.created"
    coupon_updated   = f"coupon.updated"
    coupon_deleted   = f"coupon.deleted"
    customer_created = f"customer.created"
    customer_updated = f"customer.updated"
    customer_deleted = f"customer.deleted"
    order_created    = f"order.created"
    order_updated    = f"order.updated"
    order_deleted    = f"order.deleted"
    product_created  = f"product.created"
    product_updated  = f"product.updated"
    product_deleted  = f"product.deleted"


class WooCommerceEventHandler(cbra.Endpoint):
    #default_renderer: type[cbra.renderers.IRenderer] = cbra.renderers.JSONRenderer
    method: str = "POST"
    model: types.UnionType = WooCommercePing | WooCommerceOrder
    require_authentication: bool = False
    topic: str | None
    source: str | None
    parsers: list[type[cbra.parsers.IParser]] = [
        cbra.parsers.JSONParser,
        cbra.parsers.FormParser
    ]
    renderers: list[type[cbra.renderers.IRenderer]] = [
        cbra.renderers.JSONRenderer
    ]

    @staticmethod
    async def principal_factory(
        request: cbra.Request,
        webhook_id: int,
        signature: str = cbra.Header(
            default=None,
            alias="X-WC-Webhook-Signature",
            title="Signature",
            description=(
                "A base64 encoded HMAC-SHA256 hash of the "
                "request payload."
            )
        ),
    ) -> None:
        body = await request.body()
        if not signature:
            return None
        raise NotImplementedError(body)

    def __init__(
        self,
        topic: WooCommerceTopicType | None = fastapi.Header(
            default=None,
            alias="X-WC-Webhook-Topic",
            title="Webhook topic"
        ),
        source: str | None = fastapi.Header(
            default=None,
            alias="X-WC-Webhook-Source",
            title="Source",
            example="https://example.com/"
        )
    ) -> None:
        self.source = source
        self.topic = topic
        if self.topic is not None:
            self.resource_type, self.event_type = str.split(self.topic, '.')

    async def handle(self, dto: model) -> dict[str, typing.Any]:
        return await self.discriminate(dto)

    @functools.singledispatchmethod
    def discriminate(self, dto: typing.Any) -> typing.Any:
        raise NotImplementedError

    @discriminate.register
    async def handle_ping(self, dto: WooCommercePing):
        return "PONG"

    @discriminate.register
    async def handle_order(self, dto: WooCommerceOrder):
        pass


app = cbra.Application()
app.add(WooCommerceEventHandler, base_path="/webhooks/{webhook_id}")


if __name__ == '__main__':
    cbra.run('__main__:app', reload=True)