# pylint: skip-file
import pytest
import pytest_asyncio
import ckms
from ckms.jose import JSONWebKey
from ckms.jose import JSONWebToken
from ckms.jose import PayloadCodec
from ckms.jose.exceptions import JOSEException
from cryptography.hazmat.primitives.asymmetric import ed448

from cbra.ext import oauth2


@pytest.fixture(scope='session')
def keychain():
    return ckms.Keychain()


@pytest.fixture
def private_key():
    return ed448.Ed448PrivateKey.generate()


@pytest_asyncio.fixture
async def jwk(keychain, private_key):
    keychain.register({
        'sig': {
            'provider': "local",
            'key': private_key
        }
    })
    await keychain.setup()
    spec = keychain.get('sig')
    return JSONWebKey.fromimpl(spec.key.public_key(), kid=spec.public_kid)


@pytest.fixture
def codec(keychain):
    return PayloadCodec(
        verifier=keychain,
        signer=keychain
    )


def test_allowed_redirect_urls_are_parsed_from_strings():
    config = oauth2.ClientConfig(
        client_id="foo",
        redirect_uris=["https://www.example.com"]
    )
    assert isinstance(config.redirect_uris, list)
    assert isinstance(config.redirect_uris[0], oauth2.RedirectURL)
    assert str(config.redirect_uris[0]) == "https://www.example.com"


def test_default_redirect_url_is_parsed_from_string():
    config = oauth2.ClientConfig(
        client_id="foo",
        default_redirect_url="https://www.example.com"
    )
    assert isinstance(config.default_redirect_url, oauth2.RedirectURL)


def test_can_redirect_accepts_only_whitelisted_urls():
    config = oauth2.ClientConfig(
        client_id="foo",
        redirect_uris=["https://www.example.com"]
    )
    assert config.can_redirect("https://www.example.com")
    assert not config.can_redirect("https://www.example.net")


def test_can_redirect_accepts_default():
    config = oauth2.ClientConfig(
        client_id="foo",
        default_redirect_url="https://www.example.com"
    )
    assert config.can_redirect("https://www.example.com")
    assert not config.can_redirect("https://www.example.net")


def test_get_redirect_url_with_whitelisted_url():
    config = oauth2.ClientConfig(
        client_id="foo",
        redirect_uris=["https://www.example.com"]
    )
    url = config.get_redirect_url("https://www.example.com")
    assert url == "https://www.example.com"


def test_get_redirect_url_with_non_whitelisted_url():
    config = oauth2.ClientConfig(
        client_id="foo",
        redirect_uris=["https://www.example.com"]
    )
    with pytest.raises(oauth2.RedirectForbidden):
        url = config.get_redirect_url("https://www.example.net")


def test_default_redirect_is_added_to_allowed_redirect_urls():
    config = oauth2.ClientConfig(
        client_id="foo",
        default_redirect_url="https://www.example.com"
    )
    assert "https://www.example.com" in config.redirect_uris


def test_get_redirect_url_with_none_returns_default():
    config = oauth2.ClientConfig(
        client_id="foo",
        default_redirect_url="https://www.example.com"
    )
    url = config.get_redirect_url(None)


def test_get_redirect_url_with_whitelisted_str_returns_redirect():
    config = oauth2.ClientConfig(
        client_id="foo",
        redirect_uris=["https://www.example.net"],
        default_redirect_url="https://www.example.com"
    )
    url = config.get_redirect_url("https://www.example.net")
    assert url == "https://www.example.net"


def test_get_redirect_url_with_whitelisted_redirecturi_returns_redirect():
    config = oauth2.ClientConfig(
        client_id="foo",
        redirect_uris=["https://www.example.net"],
        default_redirect_url="https://www.example.com"
    )
    url = config.get_redirect_url(oauth2.RedirectURL.validate("https://www.example.net"))
    assert url == "https://www.example.net"


def test_get_redirect_url_with_none_and_no_default():
    config = oauth2.ClientConfig(
        client_id="foo",
        redirect_uris=["https://www.example.com"]
    )
    with pytest.raises(oauth2.MissingRedirectURL):
        config.get_redirect_url(None)


@pytest.mark.asyncio
async def test_verify_valid_jwt_with_client(jwk, codec):
    config = oauth2.ClientConfig(
        client_id="foo",
        keys=[jwk]
    )
    payload = codec.encode(
        JSONWebToken(foo=1)
    )
    jws = await payload.sign('EdDSA', 'sig')
    assert await config.verify_jwt(bytes(jws), None)


@pytest.mark.asyncio
async def test_verify_invalid_jwt_with_client_raises(jwk, codec):
    config = oauth2.ClientConfig(
        client_id="foo",
        keys=[jwk]
    )
    payload = codec.encode(
        JSONWebToken(foo=1)
    )
    jws = await payload.sign('EdDSA', 'sig')
    with pytest.raises(JOSEException):
        assert not await config.verify_jwt(bytes(jws) + b'd', None)

